#include <stdio.h>
#include <string.h>

#include "attr.h"
#include "io.h"
#include "mem.h"
#include "node.h"
#include "printPS.h"
#include "symbols.h"
#include "touch.h"

attribute attributes = 0;

#define AcdTypeStandardBit 1
#define AcdTypeDummyBit 2
#define AcdTypeDraftBit 4

typedef struct _attrComponentDesc {
  char* name;
  char* shortname;
  unsigned int mask;	/* bits representing mutual exclusive options */
  unsigned int index;
  char* code;		/* postscript code to set this attribute */
  unsigned int type;	/* compare with AcdTypeDraftBit ... */
  int used;		/* is this component used anywhere */
  void * aux;
} _attrComponentDesc, *attrComponentDesc;

/*-----------------------------------------------.
 | For example if we have 4 shapes than we use   |
 | the first two bits to represent a shape of an |
 | attribute and index is the number of a        |
 | particular shape like `circle' in that        |
 | enumeration of all four shapes                |
 `-----------------------------------------------*/

_attrComponentDesc attrComps[] = {

  /*--------.
   | SHAPES |
   `--------*/

  { "box",	 0, 0xF, 0x1, "/" NODENAME " { Box } bind def", 
    0, 0, &setBoxTP },
  { "circle",	 0, 0xF, 0x2, "/" NODENAME " { Circle } bind def",
    AcdTypeStandardBit, 0, &setCircleTP },
  { "delta",	 0, 0xF, 0x3, "/" NODENAME " { Delta } bind def",
    0, 0, &setOtherTP },
  { "nabla",	 0, 0xF, 0x4, "/" NODENAME " { Nabla } bind def",
    0, 0, &setOtherTP },
  { "diamond",	 0, 0xF, 0x5, "/" NODENAME " { Diamond } bind def",
    0, 0, &setOtherTP },
  { "invisible", 0, 0xF, 0x6, "/" NODENAME " { Invisible } bind def",
    AcdTypeDummyBit, 0, &setOtherTP },
  { "empty",     0, 0xF, 0x7, "/" NODENAME " { Empty } bind def",
    0, 0, &setOtherTP },
  { "dot",	 0, 0xF, 0x8, "/" NODENAME " { Dot } bind def",
    AcdTypeDraftBit, 0, &setOtherTP },


  /*------------.
   | LINESTYLES |
   `------------*/

  { "dashed",	 0, 0xF0, 0x10, "/LS { [2 LT] 0 } bind def", 0, 0 },
  { "dotted",	 0, 0xF0, 0x20, "/LS { [1 LT 2 LT] 0 } bind def", 0, 0 },
  { "solid",	 0, 0xF0, 0x30, "/LS { [] 0 } bind def",
    AcdTypeStandardBit | AcdTypeDummyBit | AcdTypeDraftBit, 0 },


  /*---------.
   | COLOURS |
   `---------*/

  { "black",	 0, 0xF00, 0x100, "/CL { 0 0 0 } bind def",
    AcdTypeStandardBit | AcdTypeDummyBit | AcdTypeDraftBit, 0 },
  { "red",	 0, 0xF00, 0x200, "/CL { 1 0 0 } bind def", 0, 0 },
  { "green",	 0, 0xF00, 0x300, "/CL { 0 1 0 } bind def", 0, 0 },
  { "blue",	 0, 0xF00, 0x400, "/CL { 0 0 1 } bind def", 0, 0 },
  { "yellow",	 0, 0xF00, 0x500, "/CL { 1 1 0 } bind def", 0, 0 },


  /*---------------.
   | LINETHICKNESS |
   `---------------*/

  { "thick",	 0, 0x3000, 0x1000, "/LT { 1 pt mul } bind def", 0, 0 },
  { "thin",	 0, 0x3000, 0x2000, "/LT { 0.5 pt mul } bind def",
    AcdTypeStandardBit | AcdTypeDummyBit | AcdTypeDraftBit, 0 },


  /*-----------.
   | DIRECTION |
   `-----------*/

#define AcdDirectionMask (0xC000)
#define AcdForwardIndex  (0x4000)
#define AcdBackwardIndex (0x8000)

  { "forward",     0, AcdDirectionMask, AcdForwardIndex,
                          "/L { LineWithTickToCircle } bind def",
    AcdTypeStandardBit, 0 },

  { "backward",    0, AcdDirectionMask, AcdBackwardIndex,
                 "/L { 4 2 roll LineWithTickToCircle } bind def", 0, 0 },

  { "nondirected", 0, AcdDirectionMask, 0xC000,
                                          "/L { Line } bind def",
    AcdTypeDraftBit | AcdTypeDummyBit, 0 },

  { 0, 0, 0, 0, 0, 0, 0 } /* sentinel */
};

static int isDummyAcd(attrComponentDesc a)
  { return a->type & AcdTypeDummyBit; }

static int isStandardAcd(attrComponentDesc a)
  { return a->type & AcdTypeStandardBit; }

static int isDraftAcd(attrComponentDesc a)
  { return a->type & AcdTypeDraftBit; }

static void makeAcdNonStandard(attrComponentDesc a)
  { a->type &= ~AcdTypeStandardBit; }

static void makeAcdStandard(attrComponentDesc a)
  { a->type |= AcdTypeStandardBit; }

CellCodeTemplate(attribute)

attribute newAttr(char*s)
{
  attribute t = attribute_newCell();
  t->name = s;
  t->shortname = 0;
  t->cs = 0;
  t->brother = 0; 	/* no brother yet */
  t->father = 0; 	/* no father yet */
  t->mask = 0;
  t->TP = &setOtherTP;
  cdr(t) = attributes;
  attributes = t;

  return t;
}

static int numberOfNonStandardCS(list l)
{
  int res=0;

  for(; l; l = cdr(l))
    if(!isStandardAcd((attrComponentDesc)car(l)))
      res++;

  return res;
}

attribute dummy_attr = 0;
attribute draft_attr = 0;

void printAttribute(attribute a)
{
  list l = a->cs;
  attrComponentDesc acd;

  if( a != dummy_attr && a != draft_attr &&
      (a == standard_attr || numberOfNonStandardCS(a->cs)) )
    {
      prS(a->name);
      prS(" { ");
      while(l)
        {
          acd = (attrComponentDesc)car(l);
          if(!isStandardAcd(acd) || a==standard_attr)
            {
              prS(acd->name);
              prC(' ');
	    }
          l = cdr(l);
        };
      prS("}\n");
    }
}

void printAllAttributes()
{
  attribute l = attributes;

  while(l)
    {
      printAttribute(l);
      l = cdr(l);
    }
}

/*---------------------------------------------------------------------------.
 | standard_attr is a template for all attributes this means if one          |
 | component of an attribute is not defined then the standard attribute of   |
 | this component is used. The user is allowed to redefine standard_attr!    |
 `---------------------------------------------------------------------------*/
attribute standard_attr = 0;

/*---------------------------------------------------------------------------.
 | if a node or an edge has no attribute then this default_attr is used      |
 `---------------------------------------------------------------------------*/
attribute default_attr = 0;


attrComponentDesc findComponentDesc(char*s)
{
  attrComponentDesc p;
  for(p=attrComps; p->name; p++)
    if(strcasecmp(p->name,s)==0)
      return p;
  return 0;
}

int isValidAttrComponent(char*s) { return findComponentDesc(s) != 0; }

static unsigned int checkCS(list l)
{
  unsigned int res=0;
  attrComponentDesc acd;

  for(; l; l=cdr(l))
    {
      acd = (attrComponentDesc)car(l);
      if(res&acd->mask)
        error("multiple defined attribute component");
      res |= acd->index;
      acd->used = 1;
    }
  return res;
}

static unsigned int insertStandardParts(unsigned int t, list* l)
{
  list p = *l;
  attrComponentDesc acd;

  for(acd=attrComps; acd->name; acd++ )
    if(isStandardAcd(acd) && !(t&(acd->mask)))
      {
        p = list_cons(acd,p);
        acd->used = 1;
	t |= acd->index;
      }
  *l = p;
  return t;
}

static void insertAllStandardParts()
{
  unsigned int t;
  attribute a;

  for(a=attributes; a; a = cdr(a))
    {
      t = checkCS(a->cs);
      a->mask = insertStandardParts(t, &a->cs);
    }
}

static void insertDummyAndDraftParts()
{
  list p = dummy_attr->cs;

  attrComponentDesc acd;
  unsigned int t = checkCS(dummy_attr->cs);

  for(acd=attrComps; acd->name; acd++)
    if(isDummyAcd(acd) && !(t&(acd->mask)))
      {
        p = list_cons(acd,p);
	acd->used = 1;
	t |= acd->index;
      }
  dummy_attr->cs = p;

  p = draft_attr->cs;
  t = checkCS(draft_attr->cs);

  for(acd=attrComps; acd->name; acd++)
    if(isDraftAcd(acd) && !(t&(acd->mask)))
      {
        p = list_cons(acd,p);
	acd->used = 1;
	t |= acd->index;
      }
  draft_attr->cs = p;
}

void insertComponents(list l, attribute a)
{
  list p=l;

  for(;p;p=cdr(p))
    car(p) = findComponentDesc((char*)car(p));
  a->cs = list_append(l,a->cs);
}

static void redefineStandardCS(list l)
{
  attrComponentDesc acd;
  for(acd=attrComps; acd->name; acd++)
    makeAcdNonStandard(acd);
  for(; l; l = cdr(l))
    makeAcdStandard((attrComponentDesc)car(l));
}

static void initStandardAttribute()
{
  unsigned int t;
  
  t = checkCS(standard_attr->cs);
  /*-----------------------------------------.
   | the build in standard is a template for |
   | userdefined standard attributes         |
   `-----------------------------------------*/
  insertStandardParts(t,&standard_attr->cs);
  redefineStandardCS(standard_attr->cs);
}

static void insertTPs()
{
  attribute a;
  for(a=attributes; a; a=cdr(a))
    {
      list cs;
      for(cs=a->cs; cs; cs=cdr(cs))
        {
	  attrComponentDesc acd = (attrComponentDesc)car(cs);
	  if(acd->aux)
	    {
	      a->TP = (void(*)(node,node,float*,float*))acd->aux;
	      break;
	    }
	}
      if(!cs) error("internal: no TP found");
    }
}

void completeAttrs()
{
  initStandardAttribute();
  insertDummyAndDraftParts();
  insertAllStandardParts();
  insertTPs();
}

void initAttributes()
{
  default_attr = findAttr("default");
  standard_attr = findAttr("standard");

  dummy_attr = findAttr(DUMMY_NODE_NAME);	/* internal */
  draft_attr = findAttr(DRAFT_ATTR_NAME);
}

static list _revDirAcd = 0;

void disposeAttributes()
{
  attribute a, tmp;
  attrComponentDesc acd;

  list_freeList(_revDirAcd);

  for(a=attributes; a; a=tmp)
    {
      tmp=a->next;
      if(a -> shortname) free(a -> shortname);
      list_freeList(a -> cs);
      freeCell(a, sizeof(_attribute));
    }
  
  for(acd=attrComps; acd -> name; acd++)
    if(acd -> shortname) free(acd -> shortname);
}

static attrComponentDesc flipAcd(attrComponentDesc d)
{
  list l; 
  attrComponentDesc p, f=0, b=0, n=0;

  if(!_revDirAcd)
    {
      for(p=attrComps; p->name; p++)
        if(strcmp(p->name,"forward")==0)
	  f = p;
	else
	if(strcmp(p->name,"backward")==0)
	  b = p;
	else
	if(strcmp(p->name,"nondirected")==0)
	  n = p;
        
	if( !f || !b || !n )
	  error("internal: nonexistent direction description");
	
	_revDirAcd =
	  list_cons( n, list_cons( n,
	  list_cons( b, list_cons( f,
	  list_cons( f, list_cons( b, 0))))));
    }
  
  for(l=_revDirAcd; l; l=cdr(cdr(l)))
    if( (attrComponentDesc)car(l) == d )
      return (attrComponentDesc)car(cdr(l));

  return d;
}

/*---------------------------------------------------------------------------.
 | the following function returns an attribute that has the same components  |
 | as the argument but has its direction flipped                             |
 `---------------------------------------------------------------------------*/
#define BROTHERSTR "-brother"

attribute brotherAttr(attribute a)
{
  attribute b;
  char* bn;
  list l;
  unsigned int m;

  if(a->brother) return a->brother;

  bn = (char*) malloc(strlen(a->name)+strlen(BROTHERSTR)+1);
  strcpy(bn, a->name);
  strcat(bn, BROTHERSTR);
  
  b = newAttr(bn);
  b->brother = a;
  a->brother = b;

  for(l=a->cs; l; l=cdr(l))
    {
      attrComponentDesc acd = (attrComponentDesc)car(l);
      acd = flipAcd(acd);
      acd->used = 1;
      b->cs = list_cons(acd, b->cs);
    }

  for(m=0, l=b->cs; l; l=cdr(l))
    m |= ((attrComponentDesc)car(l))->index;

  b->mask = m;
  return b;
}

void reverseAllAttributes()
{
  attribute l=0, n=attributes, tmp;
  int len=0;

  while(n)
    {
      tmp = cdr(n); cdr(n)=l;
      l=n; n=tmp;
      len++;
    }

  attributes=l;
  if(verboseFlag)
    fprintf(stderr, VERBOSESTR
      "found %d attributes (including standard and default)\n",
      len-2);	/* exclude draft and dummy attr */
}

#define ATTRPREFIX "A"

static char buf[40];

static int shortAttrIndex = 0;

char* shortAttrName(attribute a)
{
  if(a->shortname) return a->shortname;

  sprintf( buf, ATTRPREFIX "%x", shortAttrIndex++);
  return a->shortname = strcpy(malloc(strlen(buf)+1), buf);

/*
  sprintf( buf, ATTRPREFIX "%x", a->mask );
  return a->shortname = strcpy(malloc(strlen(buf)+1), buf);
*/
}

#define COMPPREFIX "C"

static char* shortCompName(attrComponentDesc acd)
{
  if(acd->shortname) return acd->shortname;

  sprintf( buf, COMPPREFIX "%04x", acd->index );
  return acd->shortname = strcpy(malloc(strlen(buf)+1), buf);
}

void useDraftAttr()
{
  node n;
  edge e;
  attrComponentDesc acd;

  for(n=nodes; n; n=cdr(n))
    n->attr=draft_attr;

  for(e=edges; e; e=cdr(e))
    e->attr=draft_attr;

  for(acd=attrComps; acd->name; acd++)
    acd->used = isDraftAcd(acd);
}

void printPSCodeOfUsedComps()
{
  attrComponentDesc acd;
  
  prS("\n% definitions of components of attributes\n");
  for( acd = attrComps; acd->name; acd++)
    if(acd->used)
      fprintf(io_output,"/%s { %s } bind def\n", shortCompName(acd), acd->code);
}


void installLeastAttr(attribute i)
{
  attribute a;

  if(!i) return; /* dummy node */
  for(a=attributes; a; a=cdr(a))
    if(i->mask == a->mask)
      {
        a->used = 1;
	i->shortname = shortAttrName(a);
	return;
      };
#if DEBUG
  error("internal: should never be here in installLeastAttr");
#endif
}

/*---------------------------------------------------------------------------.
 | two attributs with the same mask have exactly the same components and so  |
 | should have the same shortname                                            |
 `---------------------------------------------------------------------------*/
void normalizeAttrs()
{
  attribute a;
  node n;
  edge e;

  for(a=attributes; a; a=cdr(a))
    a->used = 0;

  for(e=edges; e; e=cdr(e))
    installLeastAttr(e->attr);

  for(n=nodes; n; n=cdr(n))
    installLeastAttr(n->attr);
}

void printPSAttributeDefs()
{
  attribute a;
  list l;

  normalizeAttrs();

  prS("\n% definitions of attributes\n");
  for(a=attributes; a; a=cdr(a))
    if(a->used)
      {
        fprintf(io_output,"/%s { ", shortAttrName(a));

        for(l=a->cs; l; l=cdr(l))
          fprintf(io_output,"%s ", shortCompName((attrComponentDesc)car(l)));

        prS(PCCNAME " } bind def\n");
      };
}

attrComponentDesc undirectedAcd = 0;

attrComponentDesc undirAcd()
{
  if(!undirectedAcd)
    {
      attrComponentDesc acd = attrComps;
      for(; !undirectedAcd && acd->name; acd++)
        if(strcmp(acd->name, "nondirected")==0)
	  undirectedAcd = acd;
    }
  return undirectedAcd;
}


/*---------------------------------------------------------------------------.
 | return an undirected version of this attribute (cf brotherAttr)           |
 `---------------------------------------------------------------------------*/
#define FATHERSTR "-father"

attribute fatherAttr(attribute a)
{
  attribute b;
  char* bn;
  list l;
  unsigned int m;

  if(a->father) return a->father;

  bn = (char*) malloc(strlen(a->name)+strlen(FATHERSTR)+1);
  strcpy(bn, a->name);
  strcat(bn, FATHERSTR);
  
  b = newAttr(bn);
  b->father = b;
  a->father = b;

  for(l=a->cs; l; l=cdr(l))
    {
      attrComponentDesc acd = (attrComponentDesc)car(l);
      if(acd->mask == AcdDirectionMask)
        acd = undirAcd();
      acd->used = 1;
      b->cs = list_cons( acd, b->cs);
    }

  for(m=0, l=b->cs; l; l=cdr(l))
    m |= ((attrComponentDesc)car(l))->index;

  b->mask = m;
  return b;
}

void correctDummyEdgesAttrs(edge f, edge l, int reverse)
{
  unsigned int m = l->attr->mask;
  if( (m & AcdForwardIndex) || (reverse && (m & AcdBackwardIndex)) )
    {
      f->attr = fatherAttr(l->attr);
    }
  else
  if( (m & AcdBackwardIndex) || (reverse && (m & AcdForwardIndex)) )
    {
      f->attr = l->attr;
      l->attr = fatherAttr(l->attr);
    }
  else
    {
      f->attr = l->attr;
    }
}
