#include <stdlib.h>

#include "io.h"
#include "layout0.h"
#include "layout1.h"
#include "mem.h"
#include "perms.h"

#include "macros.h"

static void _initDreamPositions(node*n, int num)
{
  int i;
  for(i=0; i<num; i++)
    (dreamPosition.calculate)(n[i]);
}

static void initDreamPositions(int j)
  { _initDreamPositions(levels[j].nodes,levels[j].numberOfNodes); }

/*---------------------------------------------------------------------------.
 | This is phase 2 of the Sugiyama algorithm (our phase 1 ;-)                |
 |                                                                           |
 | We order the nodes of each level according to their dreamPosition (f.e.   |
 | the barycenter of the predecessor nodes in an upward sweep and the        |
 | barycenter of the succesors in an downward sweep through the levels). So  |
 | before this procedure can be called the levels must be allocated and all  |
 | nodes have to be placed on their level. After the call all nodes should   |
 | have a valid xpos.                                                        |
 `---------------------------------------------------------------------------*/
static void phase1BodyAux(int level)
{
  qsort( levels[level].nodes, levels[level].numberOfNodes,
         sizeof(node), cmpDreamPositions);
  insertLevelIndex(level);
}

static void phase1BodyDown(int level)
{
  downwardDreamPos();
  initDreamPositions(level);
  phase1BodyAux(level);
}

static void phase1BodyUp(int level)
{
  upwardDreamPos();
  initDreamPositions(level);
  phase1BodyAux(level);
}

void sortLevelsByPredsPos()
{
  phaseBodyDown[1] = phase1BodyDown;
  phaseBodyUp[1] = phase1BodyUp;
  levelSweep(1);
}

/*---------------------------------------------------------------------------.
 | This function tests if the unplaced node n can be placed between l and r. |
 | These have to be been placed and also no placed node must lie between l   |
 | and r. If it is possible to place n an value greater than zero is         |
 | returned and *f is set to the position where n can be placed. Otherwise   |
 | zero is returned and *f is not touched. We allow l and r to be zero which |
 | represents an open intervall.                                             |
 `---------------------------------------------------------------------------*/
static int intervalDoesIt(node l, node r, node n, float* f)
{
  if(!l && !r)
    {
      *f = dreamPosition.get(n);
      return 1;
    }
  if(!l)
    {
      float rb =  r->xpos - nodeDistance(n, r);
      float dp = dreamPosition.get(n);
      if( rb <= dp )
        *f = rb;
      else
        *f = dp;
      return 1;
    }
  if(!r)
    {
      float lb = nodeDistance(l, n) + l->xpos;
      float dp = dreamPosition.get(n);
      if(lb >= dp)
        *f = lb;
      else
        *f = dp;
      return 1;
    }
  else
    {
      float
       l2n = nodeDistance(l, n),
       n2r = nodeDistance(n, r),
       lb  = l->xpos + l2n,
       rb  = r->xpos - n2r;

      if(lb<=rb)
        {
          float dp = dreamPosition.get(n);
          if(lb>dp)
            *f = lb;
          else
          if(rb<dp)
            *f = rb;
          else
            *f = dp;
          return 1;
        }
      else return 0;
    }
}

static void placeNode(node n)
{
  node tmp = levels[n->level].placedNodes;

  node left, right;
  node bestLeft;
  float bestPos;
  float bestPosDelta;

  bestLeft = 0;
  intervalDoesIt(0, tmp, n, &bestPos);
  bestPosDelta = abs(dreamPosition.get(n) - bestPos);

  right = (left = tmp) ? tmp->aux.nextPlaced : 0;
  for( ; left; )
    {
      float newPos;
      if( intervalDoesIt(left, right, n, &newPos) )
        {
	  float dp = dreamPosition.get(n);
          if( abs(newPos - dp) < bestPosDelta )
            {
              bestLeft = left;
              bestPos = newPos;
              bestPosDelta = abs(dp - bestPos);
            } 
        }
      if((left=right)) right=right->aux.nextPlaced;
    }
  
  if(!bestLeft)
    {
      n->aux.nextPlaced = levels[n->level].placedNodes;
      levels[n->level].placedNodes = n;
    }
  else
    {
      n->aux.nextPlaced = bestLeft->aux.nextPlaced;
      bestLeft->aux.nextPlaced = n;
    }
  
  n->xpos = bestPos;
}

static void resetPlacedNodes(int level)
{
  int i, max = levels[level].numberOfNodes;
  node *n = levels[level].nodes;

  for( i=0; i<max; i++)
    n[i]->aux.nextPlaced=0;

  levels[level].placedNodes = 0;
}

/*---------------------------------------------------------------------------.
 | compare two nodes according to their calculated priority                  |
 `---------------------------------------------------------------------------*/
static int priorityCmp(const void* a, const void* b)
{
  node na = *(node*)a, nb = *(node*)b;
  int pa = na->aux.priority, pb = nb->aux.priority;

  if(pa!=pb) return pa<pb ? 1 : -1;
  else       return 0;
}

static node* currentLevelNodes = 0;
static int indirectPriorityCmp(const void* a, const void* b)
{
  int ia = *(int*)a, ib = *(int*)b;
  return priorityCmp(&currentLevelNodes[ia], &currentLevelNodes[ib]);
}

/*---------------------------------------------------------------------------.
 | compare two nodes according to their xpos                                 |
 `---------------------------------------------------------------------------*/
static int xposCmp(const void* a, const void* b)
{
  node na = *(node*)a, nb = *(node*)b;
  float xa = na->xpos, xb = nb->xpos;

  if(xa!=xb) return xa<xb ? -1 : 1;
  else       return 0;
}

static void initPriorities(int level)
{
  node * a = levels[level].nodes;
  int i, num = levels[level].numberOfNodes;

  for(i=0; i<num; i++)
    a[i]->aux.priority = priorityOfNode(a[i]);
}

/*---------------------------------------------------------------------------.
 | This function places the node in one level and tries to minimize the      |
 | width for the whole level which may result in an reordering of the nodes. |
 `---------------------------------------------------------------------------*/
static void compactPlaceLevel(int level)
{
  int i, num=levels[level].numberOfNodes;
  node* a = levels[level].nodes;

  initPriorities(level);
  qsort(a, num, sizeof(node), priorityCmp); 
  resetPlacedNodes(level);

  initDreamPositions(level);
  for(i=0; i<num; i++)
    placeNode(a[i]);

  qsort(a, num, sizeof(node), xposCmp);
}

/*---------------------------------------------------------------------------.
 | This function places the nodes in one level and respects the relative     |
 | order of the nodes. This is the original Sugiyama placing function.       |
 `---------------------------------------------------------------------------*/

static void placeLevelRespectingOrder(int level)
{
#ifdef DEBUG
  int DEBUG_i;
#endif
  int i;
  node* a = currentLevelNodes = levels[level].nodes;
  const int num = levels[level].numberOfNodes;

  for(i=0; i<num; i++)
    { lBuf[i].asInt = i; a[i]->mark = 0; } /* all placed nodes are marked */

  initPriorities(level);
  qsort(lBuf, num, sizeof(lBufData), indirectPriorityCmp);
  /*-----------------------------------------------------------------*\
   | Now lBuf contains the indices of the nodes of the current level |
   | with decreasing priority.                                       |
  \*-----------------------------------------------------------------*/

  initDreamPositions(level);
  for(i=0; i<num;)
    {
      
      float dp = dreamPosition.get(a[lBuf[i].asInt]);
      int l, r, j, k, ss, es;
      float w, rb=0, lb=0;
      float pos;

      for(j=i+1; j < num; j++)
        {
	   float dp2 = dreamPosition.get(a[lBuf[j].asInt]);
           if( lBuf[j].asInt != lBuf[j-1].asInt + 1 || abs(dp2-dp)>epsilon )
	     break;
	}
      j -= i;			/* l = length of sequence ... */
      ss = lBuf[i].asInt;	/* start of sequence */
      es = ss + j;		/* end of sequence */

      /*------------------------------------------------------------------*\
       | Now we have that the sequence                                    |
       |                                                                  |
       |    (a[ lBuf[i] ], a[ lBuf[i] + 1], ... , a[ lBuf[i] + j-1])      |
       |                                                                  |
       | is equal to the sequence                                         |
       |                                                                  |
       |    (a[ lBuf[i] ], a[ lBuf[i+1] ],  ... , a[ lBuf[i+j-1] ])       |
       |                                                                  |
       | and all nodes in these sequences have approximatly the same      |
       | dream Position.                                                  |
       |                                                                  |
       | This means we can place them in that order around their dream    |
       | position. For example without this placing at once our algorithm |
       | will produce (in a downward phase with preds dreampositions)     |
       |                                                                  |
       |           o             instead of                o              |
       |           |\            more symmetric           / \             |
       |           | \                                   /   \            |
       |           o  o                                 o     o           |
       |                                                                  |
       | because in the first the left son is placed first at his dream   |
       | position.                                                        |
      \*------------------------------------------------------------------*/

      for(w=0, k=ss; k<es-1; k++)   /* calculate width of sequence */
        w += a[k]->distance;

      for(l=ss-1; l>=0 && !a[l]->mark; l--)
        ;
      for(r=es; r<num && !a[r]->mark; r++)
        ;

      /*
      ** l = index of first placed node to the left of the sequence
      ** r = index of first placed node to the right of the sequence
      */

      if(r<num)
        for(rb=a[r]->xpos, k=r-1; k>=es-1; k--)
	  rb -= a[k]->distance;

      if(l>=0)
        for(lb=a[l]->xpos, k=l; k<ss; k++)
	  lb += a[k]->distance;

      if(l<0 && r>=num)			/* first iteration only !!! */
	pos = dp - w/2;
      else
      if(l<0)				/* sequence at left border */
	pos = rb-dp<w/2 ? rb-w : dp-w/2;
      else				/* sequence at right border */
      if(r>=num)
        pos = dp-lb<w/2 ? lb : dp-w/2;
      else
        {
	  if(rb-dp<w/2)
	    pos = rb-w;
	  else
	  if(dp-lb<w/2)
	    pos = lb;
	  else
	    pos = dp-w/2;
	}

#ifdef DEBUG
      if(l>=0 && r<num && abs(rb - lb) < w - epsilon)
        {
	  fprintf( stderr, WARNINGSTR
"internal: width w (%f) > bounds rb - lb (%f) in level %d position %d\n",
	    w, abs(rb -lb - w), level, i);
	}
#endif
      
      for(k=ss; k<es; k++)		/* place sequence from left */
        {
	  a[k]->xpos = pos;
	  a[k]->mark = 1;
	  pos += a[k]->distance;
	}

      i += j;
    }

#ifdef DEBUG
    for(DEBUG_i=0; DEBUG_i<num-1; DEBUG_i++)
      if( a[DEBUG_i]->xpos > a[DEBUG_i+1]->xpos )
        fprintf( stderr, WARNINGSTR "internal: order violated\n");
#endif
}

static void phase3and4BodyAux(int level)
{
  if(compactifyLevels)
    compactPlaceLevel(level);
  else
    placeLevelRespectingOrder(level);
}

static void phase3and4BodyDown(int level)
{
  downwardDreamPos();
  phase3and4BodyAux(level);
}

static void phase3and4BodyUp(int level)
{
  upwardDreamPos();
  phase3and4BodyAux(level);
}

/*---------------------------------------------------------------------------.
 | Phase 3 and 4 place each node according to its dream position. In the     |
 | first case only the predecessors or succesors are investigated to         |
 | calculate the dreamPositions. In the second case both, preds and succs,   |
 | are used. The difference between this phases and phase 2 is that real x   |
 | positions are calculated and not only a level ordering is generated.      |
 |                                                                           |
 | Comments:                                                                 |
 |                                                                           |
 | (1)     The original Sugiyama Algorithm suggest that phase 2 is a         |
 |         requirement for phase 3. But it seems to work if we skip phase 2  |
 |         by setting phaseIterations[2] to zero.                            |
 |                                                                           |
 | (2)     Phase 4 seems a good thing for small graphs. But for large        |
 |         graphs it makes the graph look like a fractal. Perhaps because    |
 |         almost all lines will have an slope != 0.                         |
 `---------------------------------------------------------------------------*/
void placeNodes3()
{
  useSinglePosition();
  phaseBodyDown[3] = &phase3and4BodyDown;
  phaseBodyUp[3] = &phase3and4BodyUp;
  levelSweep(3);
}

void placeNodes4()
{
  useDoublePosition();
  phaseBodyDown[4] = &phase3and4BodyDown;
  phaseBodyUp[4] = &phase3and4BodyUp;
  levelSweep(4);
}

#ifdef MAXWINDOWSIZE_IS_7
#define MAXWINDOWSIZE 7
#else
#define MAXWINDOWSIZE 5
#endif

#define MINWINDOWSIZE 2

static int maxWindowSize = MAXWINDOWSIZE;
static int minWindowSize = MINWINDOWSIZE;

static const Permutations* nodeSwapPerm = 0;
static int edgeCrossMinimizationPassesPerLevel = 1;

void setWindowSizeTo(int i)
{
  switch(i) {
    case 2: nodeSwapPerm = &perms2; break;
    case 3: nodeSwapPerm = &perms3; break;
    case 4: nodeSwapPerm = &perms4; break;
    case 5: nodeSwapPerm = &perms5; break;
#ifdef MAXWINDOWSIZE_IS_7
    case 6: nodeSwapPerm = &perms6; break;
    case 7: nodeSwapPerm = &perms7; break;
#endif
    default:
      fprintf(stderr,
WARNINGSTR "wrong windowsize -- must lie between 2 and %d\n"
WARNINGSTR "using standard size of 3 instead\n", MAXWINDOWSIZE);
      nodeSwapPerm = &perms3;
      break;
  }
}

void setECMIterationsPerLevel(int i)
  { edgeCrossMinimizationPassesPerLevel = i; }

static void _edgeCrossMinimization(int level)
{
  node tmp_nodes[MAXWINDOWSIZE];

  const unsigned int *end_of_perms = 0;
  const unsigned int *current_perm;
  const unsigned int num = levels[level].numberOfNodes;
  node *n = levels[level].nodes;

  int windowSize;
  unsigned int inc;
  int lasti, iterations = edgeCrossMinimizationPassesPerLevel;

  int swaped_something = 1;
  const Permutations* mySwapPerm;

  if(num<2) return;

  if(num<nodeSwapPerm->num)
    {
      switch(num) {
        case 2: mySwapPerm = &perms2; break;
        case 3: mySwapPerm = &perms3; break;
        case 4: mySwapPerm = &perms4; break;
	case 5: mySwapPerm = &perms5; break;

#ifdef MAXWINDOWSIZE_IS_7
        case 6: mySwapPerm = &perms6; break;
	case 7: mySwapPerm = &perms7; break;
#endif

        default: mySwapPerm = &perms2;
      }
    }
  else
    mySwapPerm = nodeSwapPerm;

  windowSize = mySwapPerm->num;
  lasti =  num - windowSize;
  inc = mySwapPerm->num;
  end_of_perms = mySwapPerm->a + mySwapPerm->rows * inc;

  for(; iterations && swaped_something; iterations--)
    {
      int i;

      for(i=0; i<=lasti; i++)
        {
          const unsigned int *minPerm;
          unsigned int minCrossings=MAXINT;
    
          int j;

	  swaped_something = 0;
    
          for(minPerm = current_perm = mySwapPerm->a;
              current_perm<end_of_perms; current_perm += inc)
            {
    	      int tmp, k;
    
    	      for(j=0; j<windowSize; j++)
                (tmp_nodes[j]=n[i+current_perm[j]])->mark = j;
    
    	      for(j=0, tmp=0; j<windowSize-1; j++)
    	        for(k=j+1; k<windowSize; k++)
                  tmp += edge_crossings(tmp_nodes[j],tmp_nodes[k]);
    	  
    	      if(tmp<minCrossings)
    	        {
    	          minCrossings = tmp;
    	          minPerm = current_perm;
    	        }
    	    }
    
          if( minPerm != mySwapPerm->identity )
            {
              for(j=0; j<windowSize; j++) tmp_nodes[j]=n[i+minPerm[j]];
              for(j=0; j<windowSize; j++) (n[i+j]=tmp_nodes[j])->mark=i+j;
	      swaped_something = 1;
    	    }
	  else
	    {
	      for(j=0; j<windowSize; j++) n[i+j]->mark=i+j;
	    }
        }
    }
}

void initEdgeCrossMinimization()
{
  int level;

  if(!nodeSwapPerm) nodeSwapPerm = &perms3;
  else
  if( nodeSwapPerm->num>maxWindowSize || nodeSwapPerm->num<minWindowSize)
    error("internal: initEdgeCrossMinimization (false windowsize)");

  phaseBodyDown[2] = phaseBodyUp[2] = &_edgeCrossMinimization;

  /*-------------------------------------------------*\
   | mark all nodes with their position in the level |
  \*-------------------------------------------------*/
  for(level=0; level<numberOfLevels; level++)
    {
      node *n = levels[level].nodes;
      int num = levels[level].numberOfNodes, i;
      for(i=0; i<num; i++)
        n[i]->mark = i;
    }
}

void edgeCrossMinimization()
{
  int ec_before=0;

  if(!phaseIterations[2]) return;

  initEdgeCrossMinimization();
  if(verboseFlag) ec_before = numberOfEdgeCrossings();
  levelSweep(2);
  if(verboseFlag)
    {
      int ec_after = numberOfEdgeCrossings();
      if(ec_after<ec_before)
        fprintf(stderr,
VERBOSESTR "reduced number of edge crossings from %d to %d (%d)\n",
          ec_before, ec_after, ec_before - ec_after);
      else
      if(ec_after == ec_before)
        fprintf(stderr,
VERBOSESTR "number of edge crossings stays the same (%d)\n",
          ec_after);
      else
        error("internal: number of edge crossings increased");
    }
}
