#include "node.h"

#define cm2inch (1/2.54)
#define inch2points (72)
#define cm2points (72/2.54)

float scale = 1;	/* only factors */
float yscale = 1;
float xscale = 1;

/*---------------------------------------------------------------------------.
 | standard Pages Specs are `a4' with one inch border at *each* side         |
 `---------------------------------------------------------------------------*/
#define xsize_std (21 * cm2points)
#define ysize_std (29.7 * cm2points)
#define xoffset_std (inch2points)
#define yoffset_std (inch2points)


float xsize = xsize_std;
float ysize = ysize_std;
float yoffset = yoffset_std;
float xoffset = xoffset_std;

float borderLeft, borderRight, borderTop, borderBottom;	/* in points */

float xshift = 0;		/* in user space units */
float yshift = 0;		/* in user space units */

char buffer[100];

const float units2pointsFactor = 20;
/*---------------------------------------------------------------------------.
 | The next functions scale down user space units into points.               |
 `---------------------------------------------------------------------------*/
float xUnits2points(float f)
{
  return f * xscale * scale * units2pointsFactor +
    xoffset + xshift + borderLeft;
}

float yUnits2points(float f)
{
  return f * yscale * scale * units2pointsFactor +
    yoffset + yshift + borderTop;
}

int nonStandardWidth=0;
int nonStandardHeight=0;
int encapsulated=0;
int landscape=0;

/*---------------------------------------------------------------------------.
 | We use floats in three metric systems. The first one is our output unit   |
 | defined as the standard distance between two nodes. The other sizes are   |
 | stored as Postscript Points which are approximatley 1/72 of an inch       |
 | (~2.54 cm) But here again we have two different scales. The `real' points |
 | are those unscaled points of a postscript device -- so to say the metrics |
 | of the raw output device. Than we use `scaled' points (scaled with        |
 | `scale'). These Points are used to define `line thickness', diameter of   |
 | Dots and Circles etc.  Here I should mention the role of `yscale' and     |
 | `xscale'. They are used two scale `units' in y (resp. x) directions. So   |
 | if you think your graph should be twice as wide then set xscale=2 and all |
 | the other features like line thickness and especially the height of the   |
 | graph will stay the same. This is also used to fit a graph to a given     |
 | page size without enlarging graph elements like circles line thickness    |
 | etc.                                                                      |
 `---------------------------------------------------------------------------*/
static void SetPageSizes(float xs, float xo, float ys, float yo)
{
  if(!encapsulated)
    {
      if(xoffset==xoffset_std) xoffset = xo;
      if(yoffset==yoffset_std) yoffset = yo;
    }
  if(!nonStandardWidth)
    {
      if(xsize==xsize_std) xsize = xs;
      if(ysize==ysize_std) ysize = ys;
    }
  if(landscape)
    {
      /* swap x and y */
      float tmp;
      tmp = xsize; xsize = ysize; ysize = tmp;
      tmp = xoffset; xoffset = yoffset; yoffset = tmp;
    }
}



void a4PageSize()
{ SetPageSizes(21 * cm2points, inch2points, 29.7 * cm2points, inch2points); }
void letterPageSize()
{ SetPageSizes(11 * inch2points, inch2points, 8 * inch2points, inch2points); } 

void (*initPageSize)() = &a4PageSize;

int centerPage=1;
int fitToPage=0, fitToGraph=0;

/*---------------------------------------------------------------------------.
 | Circle has for example a radius of 3 scaled PS points. This means it has  |
 | a 3/units2pointsFactor radius in user space units.                        |
 `---------------------------------------------------------------------------*/
float standardDistance(node n) { return 6 / units2pointsFactor; }

static float (*distance2leftBorder)(node n) = &standardDistance;
static float (*distance2rightBorder)(node n) = &standardDistance;
static float (*distance2topBorder)(node n) = &standardDistance;
static float (*distance2bottomBorder)(node n) = &standardDistance;

extern int normalizeLevelsFlag;

void calculateOffsetAndSize()
{
  node n=0;

  float xmin, xmax, ymin, ymax;		/* all in units */
  float ax, ay, dx, dy;			/* all in points */
  float xs, ys;				/* both in points */
  
  initPageSize();

  if(!nodes)
    borderLeft = borderRight = borderTop = borderBottom =
    xmin = xmax = ymin = ymax = 0;
  else
    {
      borderLeft = distance2leftBorder(n),
      borderRight = distance2rightBorder(n),
      borderTop = distance2topBorder(n),
      borderBottom = distance2bottomBorder(n);

      xmax = xmin = nodes->xpos;
      ymax = ymin = nodes->ypos;
      for(n=cdr(nodes); n; n=cdr(n))
	{
	  float
	    bl = distance2leftBorder(n),
	    br = distance2rightBorder(n),
	    bt = distance2topBorder(n),
	    bb = distance2bottomBorder(n);

	  if( xmin - borderLeft > n->xpos - bl )
	    { xmin = n->xpos; borderLeft = bl; }
	  if( xmax + borderRight < n->xpos + br )
	    { xmax = n->xpos; borderRight = br; }
	  if( ymin - borderTop > n->ypos - bt )
	    { ymin = n->ypos; borderTop = bt; }
	  if( ymax + borderBottom < n->ypos + bb )
	    { ymax = n->ypos; borderBottom = bb; }
	}
    }

  if(normalizeLevelsFlag)
    {
      for(n=nodes; n; n=cdr(n))
        { n->xpos -= xmin; n->ypos -= ymin; }

      xmax = xmax - xmin; ymax = ymax - ymin; xmin = ymin = 0;
    }
  else
    {
      for(n=nodes; n; n=cdr(n))
        n->xpos -= xmin;

      ymin = 0; xmax = xmax - xmin;
      xmin = 0;
    }

  borderLeft *= scale * units2pointsFactor;
  borderRight *= scale * units2pointsFactor;	/* convert to points */
  borderTop *= scale * units2pointsFactor;
  borderBottom *= scale * units2pointsFactor;

  xs = xsize - borderLeft - borderRight - 2*xoffset;  /* real space */
  ys = ysize - borderTop - borderBottom - 2*yoffset;

  xshift = yshift = 0;
  xmax = xUnits2points(xmax); xmin = xUnits2points(xmin);
  ymax = yUnits2points(ymax); ymin = yUnits2points(ymin);

  dx = xmax - xmin; ax = (xmax + xmin)/2;
  dy = ymax - ymin; ay = (ymax + ymin)/2;

  if(fitToPage)
    {
      if(dx==0) xshift = xsize/2 - ax; else xscale = xs / dx;
      if(dy==0) yshift = ysize/2 - ay; else yscale = ys / dy;
    }
  else
  if(fitToGraph)
    {
      xsize = borderLeft + borderRight + 2*xoffset + dx;
      ysize = borderTop + borderBottom + 2*yoffset + dy;
    }
  else
  if(centerPage)
    {
      xshift = xsize/2 - ax;
      yshift = ysize/2 - ay;
    }
  else
    {
      xshift = borderLeft + xoffset - xmin;
      yshift = borderRight + yoffset - ymin;
    }
}

